#include "pogohash.h"

// based on
// https://gist.github.com/FastPokeMapDev/c1b06b7b6903e399faa67f8e7c8037a2

#include "pogohash.conf.45.0"

uc_engine *uc;
char *code_buffer;
char  mem_debug[80] = { 0 };

const int HASHSEED = 0x61247FBF;

char  neon[] = { 
                 0x4f, 0xf4, 0x70, 0x00,  // mov.w r0, #0xf00000
                 0x01, 0xee, 0x50, 0x0f,  // mcr p15, #0x0, r0, c1, c0, #0x2
                 0xbf, 0xf3, 0x6f, 0x8f,  // isb sy 
                 0x4f, 0xf0, 0x80, 0x43,  // mov.w r3, #0x40000000
                 0xe8, 0xee, 0x10, 0x3a,  // vmsr fpexc, r3
                 0x00, 0xbf               // nop
               }; 
char  stub[] = { 
                 0x90, 0x47,              // blx r2
                 0x00, 0xbf               // nop
               };
uint32_t HashFunctionAddr = 0;

uc_hook hhc_ins;
void POGO_hook(uc_engine *uc, uint64_t address, uint32_t size, void *user_data)
{
#ifdef TRACE_DEBUG
  static int ins_counter;
  static int address_lo;
  static int address_hi;
  static int address_last = 0x2000;
  int i, cnt;

  // instruction counter
  if (address == POGO_FUNC)
  {
    address_lo = address;
    address_hi = address;
    ins_counter = 1;
  }
  else
  {
    if (address < address_lo) address_lo = address;
    if (address > address_hi) address_hi = address;
    ins_counter++;
  }

// unicorn BUG:
//
// size is always 2; so, we need to do some funky hack to show when
// the instructions are 4 bytes long.. we also need to cache the memory
// access to align it correctly.. ugly hack, but works.
//

  // bugfix for previous instructions
  cnt = (((uint64_t)address - (uint64_t)address_last));
  if ((cnt == 3) || (cnt == 4)) 
  {
    for (i=2; i<cnt; i++)
      printf("%02x ", code_buffer[address_last+i] & 0xff);
  }

  // hack to show memory read/write
  if (strlen(mem_debug) > 0) printf(mem_debug);
  strcpy(mem_debug, "");

  if ((((uint64_t)address - (uint64_t)address_last) < 0) ||
      (((uint64_t)address - (uint64_t)address_last) > 4))
  {
    printf("\n");
    printf("BRANCH DETECTED");
  }

  printf("\n");
  printf("address: 0x%08lx ", address);
  printf("opcodes: ");
  for (i=0; i<2; i++)  // size is always 2 
    printf("%02x ", code_buffer[address+i] & 0xff);

  // instruction counter
  if (address == POGO_FUNC_END)
  {
     printf("\n");
     printf("address_lo: 0x%08lx ", (uint64_t)address_lo);
     printf("address_hi: 0x%08lx ", (uint64_t)address_hi);
     printf("instructions: %d\n", ins_counter);
  }

  address_last = address;
#endif
}

uc_hook hhc_mem_rw;
bool POGO_hook_mem(uc_engine *uc, uc_mem_type type, uint64_t address, int size, int64_t value, void *user_data)
{
#ifdef TRACE_DEBUG
  // we ignore anything above 0xd0000000 (heap, stack)
  if (address < 0xd0000000) 
  {   
    switch (size)
    {
      case 4:
           if (type == UC_MEM_READ)
             sprintf(mem_debug, "\n[mem-r] len:%d at 0x%08lx val: 0x%08lx", 
                     size, address, *((uint64_t *)&code_buffer[address]));
           else
           if (type == UC_MEM_WRITE)
             sprintf(mem_debug, "\n[mem-w] len:%d at 0x%08lx val: 0x%08lx", 
                     size, address, (uint64_t)value);
           break;
    }
  }
#endif
  return false;
}

uc_err POGO_createmap(uint64_t address, uint32_t size)
{
  uint32_t chunk = address % 0x1000;
  uint64_t taddr = address - chunk;
  uint32_t tsize = ((0xFFF + chunk + size) / 0x1000) * 0x1000;

  uc_err err = uc_mem_map(uc, taddr, tsize, UC_PROT_ALL);
  if (err)
  {
    printf("Failed on uc_mem_map() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

uc_err POGO_memmap(uint64_t address, char* buffer, uint32_t size)
{
  uc_err err = uc_mem_write(uc, address, buffer, size);
  if (err)
  {
    printf("Failed on uc_mem_write() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

int POGO_init()
{
  unsigned int sp;
  uc_err err;
  FILE* fp;
  long size, i;
  int exportdata, exportaddr, cnt;
 
  // dynamically load shared library
#ifdef DYNLOAD
  if (!uc_dyn_load(NULL, 0))
  {
    printf("Error dynamically loading shared library.\n");
    printf("Please check that unicorn.dll/unicorn.so is available as well as\n");
    printf("any other dependent dll/so files.\n");
    printf("The easiest way is to place them in the same directory as this app.\n");
    return 1;
  }
#endif

  // Initialize emulator in ARM 32bit thumb mode
  err = uc_open(UC_ARCH_ARM, UC_MODE_THUMB, &uc);
  if (err)
  {
    printf("Failed on uc_open() with error returned: %u\n", err);
    return err;
  }

  // hook all instructions by having @begin > @end
  uc_hook_add(uc, &hhc_ins, UC_HOOK_CODE, (void *)POGO_hook, NULL, (uint64_t)1, (uint64_t)0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  uc_hook_add(uc, &hhc_mem_rw, UC_HOOK_MEM_READ | UC_HOOK_MEM_WRITE, POGO_hook_mem, NULL, 1, 0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }

  // heap and stack allocation
  POGO_createmap(0xE0000000, 0x2000);
  sp = 0xE0000000 + 0x2000; // stack grows down within heap

  // stack
  uc_reg_write(uc, UC_ARM_REG_SP, &sp);

  // neon support enable 
  POGO_createmap(0x1000, 0x1000);
  POGO_memmap(0x1000, neon, sizeof(neon));

  // start code calling hash func
  POGO_createmap(0x2000, 0x2000);
  POGO_memmap(0x2000, stub, sizeof(stub));

  // binary mapping
  fp = fopen(POGO_FILENAME, "rb");
  size = POGO_BIN_SIZE; 
  code_buffer = malloc(size);
  fseek(fp, 0, SEEK_SET);
  fread(code_buffer, size, 1, fp);
  fclose(fp);

  // memory prepare
  POGO_createmap(POGO_BIN_OFFSET, POGO_BIN_MAX - POGO_BIN_OFFSET);

#if (POGO_PATCH_VER == 45)
  // specific for pokemon.45.0
  {
    char patch[2] = { 0x00, 0xbf }; // NOP 
    memmove(code_buffer + 0x01B1769C, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B1769E, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B17CD4, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B17CD6, patch, sizeof(patch)); 
  }
  // specific for pokemon.45.0: end
#endif

  // include the stub and neon for debugging
  memmove(code_buffer + 0x1000, neon, sizeof(neon)); 
  memmove(code_buffer + 0x2000, stub, sizeof(stub)); 

  // export address fill with any valid address 
  exportdata = POGO_EXP_DATA;
  exportaddr = POGO_EXP_ADDR;
  cnt = 0;
  for (i = 0; i < POGO_EXP_CNT; ++i)
  {
    int value = *(int*)(code_buffer + exportaddr + (i * 4));
    if (value == 0)
    {
      memmove(code_buffer + exportaddr + (i * 4), &exportdata, 4); 
      cnt++;
    }
  }
  printf("%d patches\n", cnt);

  // all segment
  POGO_memmap(POGO_BIN_OFFSET, 
              code_buffer + POGO_BIN_OFFSET, 
              POGO_BIN_SIZE - POGO_BIN_OFFSET);

  // enable neon
  err = uc_emu_start(uc, 0x1000, 0x1000 + sizeof(neon), 0, 0);
  if (err)
  {
    printf("Failed on uc_emu_start() with error returned %u: %s\n",
      err, uc_strerror(err));
    return -1;
  }

#ifdef TRACE_DEBUG
  printf("\n");
#endif

  HashFunctionAddr = POGO_FUNC;
  return 0;
}

void POGO_quit()
{
  // free resources
  uc_close(uc);

  free(code_buffer);

  // dynamically free shared library
#ifdef DYNLOAD
  uc_dyn_free();
#endif
}

int64_t Hash(char* buffer, int size)
{
  uint32_t r0 = 0xE0001000;  // buffer
  uint32_t r1 = size;        // size
  uint32_t r2 = HashFunctionAddr + 1;
  uint64_t ret;
  uc_err err;

  uc_reg_write(uc, UC_ARM_REG_R0, &r0);
  uc_reg_write(uc, UC_ARM_REG_R1, &r1);
  uc_reg_write(uc, UC_ARM_REG_R2, &r2);

  POGO_memmap(r0, buffer, size);

  // execute code
  err = uc_emu_start(uc, 0x2000, 0x2000 + sizeof(stub), 0, 0);
  if (err)
  {
    printf("Failed on uc_emu_start() with error returned %u: %s\n",
      err, uc_strerror(err));
    return -1;
  }
#ifdef TRACE_DEBUG
  printf("\n");
#endif

  uc_reg_read(uc, UC_ARM_REG_R0, &r0);
  uc_reg_read(uc, UC_ARM_REG_R1, &r1);

  ret = (unsigned int)r1;
  ret = (ret << 32) | (unsigned int)r0;
  return (int64_t)ret;
}

int64_t Hash64(char* buffer, int size, int salt)
{
  int64_t ret;
  char* newbuff = (char*)malloc(size + 4);
  newbuff[0] = (salt >> 24) & 0xFF;
  newbuff[1] = (salt >> 16) & 0xFF;
  newbuff[2] = (salt >> 8) & 0xFF;
  newbuff[3] = (salt >> 0) & 0xFF;
  memcpy(newbuff + 4, buffer, size);
  ret = Hash(newbuff, size + 4);
  free(newbuff);

  return ret;
}

int Hash64to32(int64_t hash)
{
  return (unsigned int)(hash >> 0) ^ (unsigned int)(hash >> 32);
}

int Hash32(char* buffer, int size)
{
  int64_t ret = Hash64(buffer, size, HASHSEED);
  return Hash64to32(ret);
}
